# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from pathlib import Path

import duckdb
import numpy as np
import pyarrow as pa
import pyarrow.dataset as ds
from fastapi_pagination.api import create_page, resolve_params
from fastapi_pagination.bases import AbstractPage, AbstractParams

from pixano import transforms
from pixano.core import arrow_types
from pixano.data import models


def get_item_details(
    dataset: ds.Dataset,
    item_id: int,
    media_dir: Path,
    infer_datasets: list[ds.Dataset] = [],
) -> dict:
    """Get item details

    Args:
        dataset (ds.Dataset): Dataset
        item_id (int): Selected item ID
        media_dir (Path): Dataset media path
        infer_datasets (list[ds.Dataset], optional): List of inference datasets. Defaults to [].

    Returns:
        dict: ImageDetails features for UI
    """
    schema = dataset.schema
    item = duckdb.query(f"SELECT * FROM dataset WHERE id={item_id}").fetchone()

    field_names = [f.name for f in schema]
    item_dict = dict(zip(field_names, item))

    # TMP info for cat_ids issue
    debug_0 = False
    if debug_0:
        print("GT", item_dict["id"], len(item_dict["objects"]))
        for item in item_dict["objects"]:
            print("---", item["category_name"], item["category_id"])

    # Inference Merge
    for infer_ds in infer_datasets:
        inf_schema = infer_ds.schema
        inf_item = duckdb.query(f"SELECT * FROM infer_ds WHERE id={item_id}").fetchone()
        if inf_item is not None:
            inf_field_names = [f.name for f in inf_schema]
            inf_item_dict = dict(zip(inf_field_names, inf_item))
            item_dict["objects"].extend(inf_item_dict["objects"])
            # TMP info for cat_ids issue
            if debug_0:
                print(
                    "INFER",
                    inf_item_dict["id"],
                    inf_item_dict["objects"][0]["bbox_source"],
                    len(inf_item_dict["objects"]),
                )
                for item in inf_item_dict["objects"]:
                    print("---", item["category_name"], item["category_id"])

    # TODO compute statically
    category_ids = [obj["category_id"] for obj in item_dict["objects"]]
    category_names = [obj["category_name"] for obj in item_dict["objects"]]
    cat, index, count = np.unique(category_ids, return_index=True, return_counts=True)
    category_stats = [
        {
            "id": int(cat[i]),
            "name": str(category_names[index[i]]),
            "count": int(count[i]),
        }
        for i in range(len(cat))
    ]

    features = {
        "id": item_dict["id"],
        "filename": None,
        "width": None,
        "height": None,
        "categoryStats": category_stats,
        "views": {},
    }

    def _format_bbox(bbox, is_predicted=False, confidence=None):
        return {
            "x": bbox[0],
            "y": bbox[1],
            "width": bbox[2],
            "height": bbox[3],
            "is_predict": is_predicted,
            "confidence": confidence,
        }

    for f in schema:
        if arrow_types.is_image_type(f.type):
            bboxes = [
                _format_bbox(
                    obj["bbox"],
                    obj["bbox_confidence"] is not None,
                    obj["bbox_confidence"],
                )
                for obj in item_dict["objects"]
                if obj["view_id"] == f.name and obj["bbox"] is not None
            ]
            masks = [
                transforms.rle_to_polygons(obj["mask"])
                for obj in item_dict["objects"]
                if obj["view_id"] == f.name
            ]

            im = arrow_types.Image(**item_dict[f.name])
            im.uri_prefix = media_dir

            features["views"][f.name] = {
                "image": im.url,
                "objects": {
                    "category": [
                        {"id": id, "name": name}
                        for (id, name) in zip(category_ids, category_names)
                    ],
                    "boundingBox": bboxes,
                    "segmentation": masks,
                },
            }

    return features


def get_items(dataset: ds.Dataset, params: AbstractParams = None) -> AbstractPage:
    """Get items

    Args:
        dataset (pa.Dataset): Dataset
        params (AbstractParams, optional): FastAPI params for pagination. Defaults to None.

    Returns:
        AbstractPage: List of models.Feature for UI (DatasetExplorer)
    """

    params = resolve_params(params)
    raw_params = params.to_raw_params()

    total = dataset.count_rows()
    schema = dataset.schema

    items_table = duckdb.query(
        # HACK TO SELECT ONLY VAL FOR COCO
        # f"SELECT * from dataset WHERE split='val2017' OFFSET {raw_params.offset} LIMIT {raw_params.limit}"
        f"SELECT * from dataset OFFSET {raw_params.offset} LIMIT {raw_params.limit}"
    ).arrow()

    def _create_features(row: list) -> list[models.Feature]:
        features = []
        for field in schema:
            if arrow_types.is_number(field.type):
                features.append(
                    models.Feature(
                        name=field.name, dtype="number", value=row[field.name]
                    )
                )
            elif arrow_types.is_image_type(field.type):
                im = arrow_types.Image(**row[field.name])

                features.append(
                    models.Feature(name=field.name, dtype="image", value=im.preview_url)
                )
            # elif types.is_depth_map_type(field.type):
            #     features.append(
            #         models.Feature(
            #             name=field.name,
            #             dtype="image",
            #             value=row[field.name].display(size=128),
            #         )
            #     )
            elif pa.types.is_string(field.type):
                features.append(
                    models.Feature(name=field.name, dtype="text", value=row[field.name])
                )

        return features

    items = [_create_features(el) for el in items_table.to_pylist()]

    return create_page(items=items, total=total, params=params)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import unittest
from io import BytesIO

import pyarrow as pa
import pyarrow.parquet as pq
import requests
from PIL import Image

from pixano.core import arrow_types
from pixano.transforms import encode_png


class TestImageUriType(unittest.TestCase):
    def test_image_uri_type(self):
        uri = "http://farm3.staticflickr.com/2595/3984712091_e82c5ec1ca_z.jpg"
        im_data = requests.get(uri)
        im = Image.open(BytesIO(im_data.content))
        im.thumbnail((128, 128))
        im.save("thumb.png")
        preview = encode_png(im)

        im_uri_type = arrow_types.ImageUriType()
        im_storage = pa.array([{"preview": preview, "uri": uri}])
        arr = pa.ExtensionArray.from_storage(im_uri_type, im_storage)

        schema = pa.schema(
            [
                pa.field("image", arrow_types.ImageUriType()),
            ]
        )
        table = pa.Table.from_arrays([arr], schema=schema)
        pq.write_table(table, "test.parquet", store_schema=True)

        print(table.schema)
        # self.assertTrue(isinstance(im_py, arrow_types.ImageUri))
        # self.assertEqual(im_py.uri, uri)
        # with open("test.jpg", "wb") as f:
        #     f.write(im_py.preview)
        # self.assertFalse(isinstance(im_uri_type, arrow_types.ImageUriType))



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import io
import json
import shutil
import uuid
from pathlib import Path

import pyarrow as pa
import pyarrow.dataset as ds
from PIL import Image
from tqdm import tqdm

from pixano.core import arrow_types


def media_copy(
    images_path: Path, filename: str, path: Path, view: str, multiview: bool
):
    """Copy file from source to media directory

    Args:
        images_path (Path): Image path
        filename (str): Image filename
        path (Path): Dataset path
        view (str): View name
        multiview (bool): True if multiview

    Raises:
        Exception: Empty filename
        Exception: File not found
        Exception: File copy error
    """

    # Check provided filename
    if filename is None or len(filename) == 0:
        raise Exception("No uri given")

    # Get file
    file = Path(images_path) / Path(filename)

    # Check if file exists
    if not file.is_file():
        # Because of inconsistent old Pixano format, try this first
        file = Path(images_path) / Path(filename).name
        if not file.is_file():
            raise Exception(f"File not found (path: {images_path}, filename: {file})")

    # Media path
    media_path = path / "media"
    if multiview:
        media_path = media_path / view
    media_path.mkdir(parents=True, exist_ok=True)

    # Copy
    try:
        shutil.copy(file, media_path / file.name)
    except Exception as e:
        raise Exception(
            f"Image copy error: {e} (src: {file}, dst: {str(media_path / file.name)})"
        )


def generate_spec(
    split_info: dict,
    path: Path,
    name: str,
    description: str = "",
    limits: list[int] = [],
):
    """Generate spec.json

    Args:
        split_info (dict): Generator for each split and views
        path (Path): Dataset path
        name (str): Dataset name
        description (str, optional): Dataset description. Defaults to "".
        limits (list[int], optional): Image limits per split. Defaults to [].

    Raises:
        Exception: File creation error
    """

    spec = {}
    spec["id"] = uuid.uuid4().hex
    spec["name"] = str(name)
    spec["description"] = description if description else str(name)

    if limits:
        spec["num_elements"] = sum(limits)
    else:
        num_elements = 0
        for split in split_info:
            if hasattr(split_info[split], "info"):
                num_elements += split_info[split].info["nb_images"]
            else:
                counts = [d.info["nb_images"] for d in split_info[split].values()]
                if counts.count(counts[0]) != len(counts):
                    raise Exception(
                        "Images counts are not consistent across views:"
                        + ", ".join(str(k) for k in split_info[split])
                    )
                num_elements += counts[0]
        spec["num_elements"] = num_elements

    try:
        with open(path / "spec.json", "w") as f:
            json.dump(spec, f)
            print("File " + str(path) + "/spec.json written")
    except IOError as err:
        raise Exception(f"Error creating spec.json file: {err}")


def generate_parquet(
    split_info: dict, path: Path, schema: pa.schema = None, limits: list[int] = []
):
    """Generate parquet file

    Args:
        split_info (dict): Generators for each split and view
        path (Path): Dataset path
        schema (pa.schema, optional): Dataset PyArrow schema. Defaults to None.
        limits (list[int], optional): Image limits per split. Defaults to None.
    """

    db_path = path / "db"
    db_path.mkdir(parents=True, exist_ok=True)

    for i, split in enumerate(split_info):
        split_limit = limits[i] if limits else 0
        lfeat = []
        if hasattr(split_info[split], "info"):
            # SingleView
            multiview = False
            datas = [split_info[split]]
            views = [
                "image"
            ]  # WARNING: We should try to get "only view" name ("image" by default, but could be "rgb")
            nb_images = datas[0].info["nb_images"]
        else:
            # MultiView: one view per generator
            multiview = True
            datas = split_info[split].values()
            views = list(split_info[split].keys())
            counts = [d.info["nb_images"] for d in datas]
            if counts.count(counts[0]) != len(counts):
                raise Exception(
                    "Images counts are not consistent across views:"
                    + ", ".join(str(v) for v in views)
                )
            nb_images = counts[0]

        try:
            if nb_images:
                if split_limit is 0 or split_limit > nb_images:
                    split_limit = nb_images
                    print(
                        f"Reading data for whole dataset (split:{split}) of {split_limit} items."
                    )
                else:
                    print(
                        f"Reading data for subset dataset (split:{split}) of first {split_limit}/{nb_images} items."
                    )
                for i in tqdm(range(split_limit)):
                    feat = {}
                    for i, data in enumerate(datas):
                        view_feat = next(data)
                        if "uri" in view_feat[views[i]]:
                            media_copy(
                                data.info["images_path"],
                                view_feat[views[i]]["uri"],
                                path,
                                views[i],
                                multiview,
                            )

                        if "objects" in feat:
                            # extend annotations
                            view_feat["objects"].extend(feat["objects"])
                        feat.update(view_feat)
                    feat["split"] = split
                    lfeat.append(feat)
            else:
                # TODO not consistent anymore with if case...
                print(
                    "Generating parquet file. Number of image unknown. Prints ./* every 10/100 items."
                )
                i = 0
                # TODO multiview
                while (feat := next(data)) and (split_limit < 0 or i < split_limit):
                    feat["split"] = split
                    lfeat.append(feat)
                    i = 1 + i
                    if i % 100 == 0:
                        print("*", end="", flush=True)
                    else:
                        if i % 10 == 0:
                            print(".", end="", flush=True)
                    if i == split_limit:
                        raise StopIteration
        except StopIteration:
            print("Done!")

        if schema is None:
            fields = [
                pa.field("id", pa.string()),
                pa.field("objects", pa.list_(arrow_types.ObjectAnnotationType())),
                pa.field("split", pa.string()),
            ]
            for view in views:
                fields.append(pa.field(view + ".width", pa.int32(), nullable=True))
                fields.append(pa.field(view + ".height", pa.int32(), nullable=True))
                fields.append(pa.field(view, arrow_types.ImageType()))

            schema = pa.schema(fields)

        # Transpose lfeats (list of rows to cols (list of col))
        cols = {k: [dic[k] for dic in lfeat] for k in lfeat[0]}

        arrays = []
        for field in schema:
            if field.name in cols:
                col = cols[field.name]
                arr = arrow_types.convert_field(field.name, field.type, col)
                arrays.append(arr)
            else:
                print("Incorrect field:", field.name)

        table = pa.Table.from_arrays(arrays=arrays, schema=schema)

        ds.write_dataset(
            data=table,
            base_dir=db_path / split,
            basename_template=f"part-{{i}}.parquet",
            format="parquet",
            max_rows_per_file=2048,
            max_rows_per_group=2048,
            existing_data_behavior="overwrite_or_ignore",
        )

        print("Parquet written in " + str(db_path / split))

    # PREVIEW - create file "preview.png"
    # get 6 first (todo? random) images
    # get image field names for preview
    image_fields = []
    for f in schema:
        if arrow_types.is_image_type(f.type):
            image_fields.append(f.name)

    if len(image_fields) > 0:
        tile_w = 64
        tile_h = 64
        preview = Image.new("RGB", (3 * tile_w, 2 * tile_h))
        for i, f in enumerate(lfeat[:6]):
            im = f[image_fields[i % len(image_fields)]]
            image = Image.open(io.BytesIO(im["preview_bytes"]))
            preview.paste(image, ((i % 3) * tile_w, (int(i / 3) % 2) * tile_h))
        preview.save(path / "preview.png")


def convert(
    split_info: dict,
    library_path: str,
    name: str,
    schema: pa.schema = None,
    limits: list[int] = [],
    description: str = "",
):
    """Create Pixano parquet dataset from generator(s)

    Args:
        split_info (dict): Generators for each split and view
        library_path (str): Dataset library path
        name (str): Dataset name
        schema (pa.schema, optional): Dataset PyArrow schema. Defaults to None.
        limits (list[int], optional): Image limits per split. Defaults to [].
        description (str, optional): Dataset description. Defaults to "".
    """

    path = Path(library_path) / name
    generate_parquet(split_info, path, schema, limits)
    generate_spec(split_info, path, name, description, limits)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import os

from PIL import Image
from pycocotools.coco import COCO

from pixano import transforms
from pixano.core import arrow_types

from .data_loader import DataLoader


class CocoLoader(DataLoader):
    """COCO Data Loader

    Attributes:
        coco (COCO): COCO dataset
        info (dict): Dataset info
        img_ids (list[int]): Image IDs
        iter_img_ids (iter): Image IDs iterable
        idx (int): Index
    """

    def __init__(self, workspace: str, ann_file: str, img_path: str):
        """Initialize COCO Data Loader

        Args:
            workspace (str): Data path
            ann_file (str): Annotations path
            img_path (str): Images path
        """

        annf = os.path.join(workspace, ann_file)

        # initialize COCO api for instance annotations
        self.coco = COCO(annf)

        self.img_ids = self.coco.getImgIds()
        self.img_ids.sort()
        self.info = self.coco.dataset["info"]
        self.info["annotation_file"] = annf
        self.info["images_path"] = os.path.join(workspace, img_path)
        self.info["nb_images"] = len(self.img_ids)

        # make img_ids iterable
        self.iter_img_ids = iter(self.img_ids)
        self.idx = 0

    def load_ann(self, id: int, width: int, height: int) -> list[dict]:
        """Load COCO annotations

        Args:
            id (int): image id
            width (int): image width, for normalization
            height (int): image height, for normalization

        Returns:
            list[dict]: Annotation data
        """

        ann_ids = self.coco.getAnnIds(id)
        anns = self.coco.loadAnns(ids=ann_ids)
        objects = []
        for ann in anns:
            bbox = None
            if ann["bbox"]:
                bbox = transforms.normalize(ann["bbox"], width, height)
            rle = None
            if ann["segmentation"]:
                rle = self.coco.annToRLE(ann)

            # TMP to check
            if isinstance(rle, list) and len(rle) > 1:
                print("WARNING - MULTI RLE SPOTTED !!", len(rle), rle)

            objects.append(
                arrow_types.ObjectAnnotation(
                    id=str(ann["id"]),
                    view_id="image",
                    is_group_of=bool(ann["iscrowd"]),
                    category_id=ann["category_id"],
                    category_name=self.coco.loadCats(ann["category_id"])[0]["name"],
                    bbox=bbox,
                    mask=rle,
                ).dict()
            )

        # print("OBJTYPE:", type_spec(objects))
        return objects

    def __iter__(self):
        self.idx = 0
        return self

    def __next__(self):
        id = next(self.iter_img_ids)
        ex = self.coco.loadImgs(ids=id)[0]
        image = Image.open(os.path.join(self.info["images_path"], ex["file_name"]))
        anns = self.load_ann(id, image.width, image.height)

        # thumbnail -- after load_ann else image.width/.height are thumbnail height/width
        image.thumbnail((128, 128))
        image_thumb = transforms.image_to_binary(image, "JPEG")

        # tmp, previous col names for now
        row = {
            "id": str(id),
            "image": {
                "uri": ex["file_name"],
                "bytes": None,
                "preview_bytes": image_thumb,
            },
            "image.width": image.width,
            "image.height": image.height,
            "objects": anns,
        }

        return row



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import html
import json

import IPython.display
import shortuuid


def get_env() -> str:
    """Get current running environment

    Returns:
        str: Running environment
    """

    # If Google colab import succeeds
    try:
        import google.colab
        import IPython
    except ImportError:
        pass
    else:
        if IPython.get_ipython() is not None:
            return "colab"

    # If IPython import succeeds
    try:
        import IPython
    except ImportError:
        pass
    else:
        ipython = IPython.get_ipython()
        if ipython is not None and ipython.has_trait("kernel"):
            return "ipython"

    # Else
    return "none"


def display(url: str = "http://127.0.0.1", port: int = 8000, height: int = 1000):
    """Display a Pixano app inside an environment

    Args:
        url (str, optional): Pixano app URL. Defaults to "127.0.0.1".
        port (int, optional): Pixano app port. Defaults to 8000.
        height (int, optional): Frame height. Defaults to 1000.
    """

    # Select function based on environment
    f = {
        "colab": display_colab,
        "ipython": display_ipython,
        "none": display_cli,
    }[get_env()]

    # Display Pixano app
    f(url, port, height)


def display_colab(url: str, port: int, height: int):
    """Display a Pixano app inside a Google Colab

    Args:
        url (str): Pixano app URL
        port (int): Pixano app port
        height (int): Frame height
    """

    # Define frame template
    shell = """
        (async () => {
            const url = new URL(%URL%);
            const port = %PORT%;
            if (port) {
                url.port = port;
            }
            const iframe = document.createElement('iframe');
            iframe.src = url;
            iframe.setAttribute('width', '100%');
            iframe.setAttribute('height', '%HEIGHT%');
            iframe.setAttribute('frameborder', 0);
            document.body.appendChild(iframe);
        })();
    """

    # Replace variables in template
    replacements = [
        ("%HEIGHT%", "%d" % height),
        ("%PORT%", "%d" % port),
        ("%URL%", json.dumps(url)),
    ]
    for k, v in replacements:
        shell = shell.replace(k, v)

    # Display frame
    script = IPython.display.Javascript(shell)
    IPython.display.display(script)


def display_ipython(url: str, port: int, height: int):
    """Display a Pixano app inside a Jupyter notebook

    Args:
        url (str): Pixano app URL
        port (int): Pixano app port
        height (int): Frame height
    """

    # Define frame template
    shell = """
        <iframe id="%HTML_ID%" width="100%" height="%HEIGHT%" frameborder="0">
        </iframe>
        <script>
            (function() {
                const frame = document.getElementById(%JSON_ID%);
                const url = new URL(%URL%, window.location);
                const port = %PORT%;
                if (port) {
                    url.port = port;
                }
                frame.src = url;
            })();
        </script>
    """

    # Replace variables in template
    frame_id = f"frame-{shortuuid.uuid()}"
    replacements = [
        ("%HTML_ID%", html.escape(frame_id, quote=True)),
        ("%JSON_ID%", json.dumps(frame_id)),
        ("%HEIGHT%", "%d" % height),
        ("%PORT%", "%d" % port),
        ("%URL%", json.dumps(url)),
    ]
    for k, v in replacements:
        shell = shell.replace(k, v)

    # Display frame
    iframe = IPython.display.HTML(shell)
    IPython.display.display(iframe)


def display_cli(url: str, port: int, height: int):
    """Display a Pixano app inside a command line interface

    Args:
        url (str): Pixano app URL
        port (int): Pixano app port
        height (int): Frame height
    """

    print(f"Please visit {url}:{port} in a web browser.")



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from abc import ABC, abstractmethod


class DataLoader(ABC):
    """Generic Data Loader"""

    @property
    def data_info(self):
        return self.info

    @data_info.setter
    def data_info(self, newinfo):
        self.info = newinfo

    @abstractmethod
    def __iter__(self):
        pass

    @abstractmethod
    def __next__(self):
        pass



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import json
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path


class InferenceModel(ABC):
    """Abstract parent class for OfflineModel and OnlineModel

    Attributes:
        name (str): Model name
        id (str, optional): Model ID
        device (str, optional): Model GPU or CPU device
        source (str, optional): Model source
        info (str, optional): Additional model info
    """

    def __init__(
        self,
        name: str,
        id: str = "",
        device: str = "",
        source: str = "",
        info: str = "",
    ) -> None:
        """Initialize model name and ID

        Args:
            name (str): Model name
            id (str, optional): Model ID. Defaults to "".
            device (str, optional): Model GPU or CPU device. Defaults to "".
            source (str, optional): Model source. Defaults to "".
            info (str, optional): Additional model info. Defaults to "".
        """

        self.name = name
        if id == "":
            self.id = f"{datetime.now().strftime('%y%m%d_%H%M%S')}_{name}"
        else:
            self.id = id
        self.device = device
        self.source = source
        self.info = info

    @abstractmethod
    def __call__():
        """Call model"""

        pass

    @abstractmethod
    def process_dataset(
        self,
        input_dir: Path,
        views: list[str],
        splits: list[str] = [],
        batch_size: int = 1,
    ) -> Path:
        """Process parquet dataset

        Args:
            input_dir (Path): Input parquet location
            views (list[str]): Dataset views
            splits (list[str], optional): Dataset splits, all if []. Defaults to [].
            batch_size (int, optional): Rows per batch. Defaults to 1.

        Returns:
            Path: Output parquet location
        """

        pass

    def save_json(
        self,
        output_dir: Path,
        filename: str,
        spec_json: dict,
        num_elements: int,
    ):
        """Save output .json

        Args:
            output_dir (Path): Output parquet location
            filename (str): Output .json filename
            spec_json (dict): Input parquet .json
            num_elements (int): Number of processed rows
        """

        # Load existing .json
        if (output_dir / f"{filename}.json").is_file():
            with open(output_dir / f"{filename}.json", "r") as f:
                output_json = json.load(f)
            output_json["num_elements"] += num_elements

        # Or create .json from scratch
        else:
            output_json = {
                "id": spec_json["id"],
                "name": spec_json["name"],
                "description": spec_json["description"],
                "num_elements": num_elements,
                "model_id": self.id,
                "model_name": self.name,
                "model_source": self.source,
                "model_info": self.info,
            }

        # Save .json
        with open(output_dir / f"{filename}.json", "w") as f:
            json.dump(output_json, f)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import time
from pathlib import Path
from typing import Optional

import pyarrow as pa
import pyarrow.dataset as ds
import pyarrow.parquet as pq
import pydantic


class DatasetInfo(pydantic.BaseModel):
    """DatasetInfo

    Attributes:
        id (str): Dataset ID
        name (str): Dataset name
        description (str): Dataset description
        num_elements (int): Number of elements in dataset
        preview (str, optional): Dataset preview
    """

    id: str
    name: str
    description: str
    num_elements: int
    preview: Optional[str]


class Dataset:
    """Dataset class

    Attributes:
        _path (Path): Dataset path
        _info (DatasetInfo): Dataset info
        _table (pa.Table): Dataset table
    """

    def __init__(self, path: Path):
        """Initialize dataset

        Args:
            path (Path): Dataset path
        """

        self._path = path
        self._info = DatasetInfo.parse_file(self._path / "spec.json")
        self._table = None

    @property
    def info(self):
        return self._info

    @property
    def media_path(self):
        return self._path / "media"

    @property
    def table(self):
        if self._table is None:
            st = time.process_time()
            self._table = pq.read_table(self._path / "db", use_legacy_dataset=False)
            et = time.process_time()
            res = et - st
            print("CPU Execution time:", res, "seconds")

        return self._table

    def load(self):
        partitioning = ds.partitioning(
            pa.schema([("split", pa.string())]), flavor="hive"
        )
        return ds.dataset(self._path / "db", partitioning=partitioning)


class InferenceDataset(Dataset):
    """Inference Dataset

    Attributes:
        _path (Path): Dataset path
        _info (DatasetInfo): Dataset info
        _table (pa.Table): Dataset table
    """

    def __init__(self, path: Path):
        self._path = path
        self._info = DatasetInfo.parse_file(self._path / "infer.json")
        self._table = None

    def load(self):
        partitioning = ds.partitioning(
            pa.schema([("split", pa.string())]), flavor="hive"
        )
        return ds.dataset(
            self._path,
            partitioning=partitioning,
            ignore_prefixes=["info", "infer.json"],
        )


class EmbeddingDataset(Dataset):
    """Embedding Dataset

    Attributes:
        _path (Path): Dataset path
        _info (DatasetInfo): Dataset info
        _table (pa.Table): Dataset table
    """

    def __init__(self, path: Path):
        self._path = path
        self._info = DatasetInfo.parse_file(self._path / "embed.json")
        self._table = None

    def load(self):
        partitioning = ds.partitioning(
            pa.schema([("split", pa.string())]), flavor="hive"
        )
        return ds.dataset(
            self._path,
            partitioning=partitioning,
            ignore_prefixes=["info", "embed.json"],
        )
